from __future__ import print_function

import sys

import numpy as np

sys.path.append('clique_mis')
from clique_mis import clique_mis_util as ut
from json import dumps


class Author:
    def __init__(self, name, comp_list=None, nodes=None):
        self.name = name
        if comp_list is None:
            self.comp_list = []
        else:
            self.comp_list = comp_list
        if nodes is None:
            self.nodes = []
        else:
            self.nodes = nodes

    def append_comp(self, comp):
        self.comp_list.append(comp)


class Comp:
    def __init__(self, comp, cliques=None):
        self.comp = comp
        if cliques is None:
            self.cliques = []
        else:
            self.cliques = cliques


class Clique:
    def __init__(self, nodes=None):
        if nodes is None:
            self.nodes = set()
        else:
            self.nodes = nodes
        self.self_references = 0


ijcai_authors = ["aziz-haris", "barron-cedeno-alberto", "choi-arthur",
                 "fazzinga-bettina", "infantes-guillaume",
                 "lanctot-marc", "lecue-freddy", "merett-geoff-v",
                 "mueen-abdullah", "pieris-andreas", "roger-gabriele",
                 "saffidine-abdallah", "samulowitz-horst", "veness-joel"]

ai10_2011_authors = ["chen-yiling", "conitzer-vincent", "daquin-matthieu",
                     "grauman-kristen", "leskovec-jure", "neill-b-daniel"]

ai10_2013_authors = ["ayanian-nora", "doshi-velez-finale", "knox-w-bradley",
                     "lee-honglak", "narodytska-nina", "procaccia-ariel",
                     "tellex-stefanie", "zohar-aviv"]


# Breadth first search to determine the whole clique which contains a node (
# initially specified in unvisited set), returns the clique(s) of the input
# node(s)
def bfs(G, visited, unvisited):
    if len(unvisited) == 0:
        return visited
    next_node = unvisited.pop()
    visited.add(next_node)
    neighbors = G.neighbors(next_node)
    unvisited_neighbors = set([x for x in neighbors if x not in visited and x
                               not in unvisited])
    bfs(G, visited, unvisited.union(unvisited_neighbors))
    return visited


# Given the comp graph G and the comp factor, output the size and number of
# occurrences of each clique
def comp_cliques(comp_graph):
    cls = []
    visited = set()
    while comp_graph.number_of_nodes() > len(visited):
        # Python 2
        # next_node = next(iter({k: v for k, v in comp_graph.node.iteritems() if k not in
        next_node = next(iter({k: v for k, v in comp_graph.node.items() if k not in
                               visited}))
        act_clique = Clique(bfs(comp_graph, set(), {next_node}))
        cls.append(act_clique)
        for node in act_clique.nodes:
            visited.add(node)
    return cls


def clique_histo(cls):
    sizes = [len(cl.nodes) for cl in cls]
    return dict((x, sizes.count(x)) for x in set(sizes))


def count_self_references(cls, cite_graph):
    for cl in cls:
        refs = 0
        for node1 in cl.nodes:
            for node2 in cl.nodes:
                if node1 != node2:
                    if cite_graph.has_predecessor(node1, node2):
                        refs += 1
        cl.self_references = refs

if __name__ == "__main__":
    authors = dict()
    for c in np.arange(0.1, 1, 0.1):
        for author_name in ijcai_authors:
            comp_graph = ut.comp_graph('../data/ijcai-2013/' + author_name, str(c))
            cite_graph = ut.cite_graph(
                '../data/ijcai-2013/' + author_name)[0]
            if author_name not in authors:
                authors[author_name] = Author(author_name, [], cite_graph.nodes())
                cls = comp_cliques(comp_graph)
            count_self_references(cls, cite_graph)
            authors[author_name].append_comp(Comp(c, cls))
        for author_name in ai10_2011_authors:
            comp_graph = ut.comp_graph('../data/ai10-2011/' + author_name, str(c))
            cite_graph = ut.cite_graph(
                '../data/ai10-2011/' + author_name)[0]
            if author_name not in authors:
                authors[author_name] = Author(author_name, [], cite_graph.nodes())
            cls = comp_cliques(comp_graph)
            count_self_references(cls, cite_graph)
            authors[author_name].append_comp(Comp(c, cls))
        for author_name in ai10_2013_authors:
            comp_graph = ut.comp_graph('../data/ai10-2013/' + author_name, str(c))
            cite_graph = ut.cite_graph(
                '../data/ai10-2013/' + author_name)[0]
            if author_name not in authors:
                authors[author_name] = Author(author_name, [], cite_graph.nodes())
            cls = comp_cliques(comp_graph)
            count_self_references(cls, cite_graph)
            authors[author_name].append_comp(Comp(c, cls))

    print("Clique size / Number of clique size occurences report:")
    for author_name, data in sorted(authors.iteritems()):
        print("Author: " + data.name + "; articles: " + str(len(data.nodes)))
        for c in data.comp_list:
            print("comp: " + str(c.comp) + "; " + str(dumps(clique_histo(
                c.cliques), sort_keys=True)))

    print("\n\n\n\nClique size / Number of citations within each clique report")
    for author_name, data in sorted(authors.iteritems()):
        print("Author: " + data.name + "; articles: " + str(len(data.nodes)))
        for c in data.comp_list:
            print("comp: " + str(c.comp) + "; " + str(dumps(
                dict((len(clique.nodes), clique.self_references) for
                     clique in c.cliques), sort_keys=True)))
