## Plot h-index increases vs. number k of operations

## setwd("~/proj/split-index/src")
library(tikzDevice)
library(ggplot2)
library(tidyverse)

source("plots_util.R")

## Read csv
dat <- read.csv("exp.csv", 
                colClasses = c("factor", "factor",
                               "factor", "factor",
                               "factor", "factor",
                               "integer", "integer",
                               "integer", "integer"))

## Nicer prints
dat <- as.tibble(dat)

dat <- latexlabels(dat)

## Replace fusioncite problem names
dat$problem <- as.factor(sub("fusion_", "", dat$problem))
    #if_else(startsWith(dat$problem, "fusion_"), substr(as.string(dat$problem), 8), dat$problem)

## Compute hindex increases
dat <- mutate(dat, hindex_increase = max_hindex - hmerged)

## We now want to restrict the data to those authors where the h-index
## could be improved.

## Add column with max hindex increases per author in each experiment
## (tuple of dataset, comp, merge_method, measure, problem).
dat <- dat %>%
    group_by(dataset, comp, merge_method, measure, problem, author) %>%
    mutate(max_increase = max(hindex_increase)) %>%
    ungroup

## Keep only data for (author, experiment) with positive max hindex
## increase.
dat <- filter(dat, max_increase > 0)

## We need to ensure that for each plot instance (corresponding to a
## tuple of dataset, comp, merge_method, problem) all authors have
## some data for each occuring value of k. We achieve this by
## repeating rows for authors whose largest value of k is smaller than
## the maximum k over all authors.

## Compute max k's over all authors, for each experiment (tuple of
## (dataset, comp, merge_method, measure, problem)).
add_maxk <- dat %>% 
    group_by(dataset, comp, merge_method, measure, problem) %>%
    mutate(maxk = max(k)) %>%
    ungroup()

## Compute max k's for each author, for each tuple of (dataset, comp,
## merge_method, measure, problem)
add_author_maxk <- add_maxk %>%
    group_by(dataset, author, comp, merge_method, measure, problem) %>%
    mutate(author_maxk = max(k)) %>%
    ungroup()

## Compute rows to repeat and how often
repeats <- add_author_maxk %>%
    group_by(dataset, author, comp, merge_method, measure, problem) %>%
    ## filter(k == author_maxk) %>%
    mutate(repeats = if_else(k == author_maxk, maxk - author_maxk + 1, 1)) %>%
    ungroup()

## Repeat rows to repeat
tidied <- repeats %>%
    uncount(repeats, .id = "uncounted") %>%
    mutate(k = k + uncounted - 1)
    ## group_by_at(vars(-k)) %>%
    ## expand(k = seq(k : (k + repeats)) + k) # wrong

print(tidied)

## Next, we want to find the largest k where for some author something
## happens, i.e., the hindex could still be improved.
changed <- tidied %>%
    group_by(dataset, author, comp, merge_method, measure, problem) %>%
    arrange(k) %>%
    mutate(changed = k == 0 | hindex_increase != lag(hindex_increase)) %>%
    ungroup()

max_changed_k <- changed %>%
    group_by(dataset, comp, merge_method, problem) %>%
    filter(changed == TRUE) %>%
    summarize(max_changed_k = max(k))

datasets <- levels(tidied$dataset)
comps <- levels(tidied$comp)
merge_methods <- levels(tidied$merge_method)
## measures <- levels(tidied$measure)
probs <- levels(tidied$problem)

## Individual plots

## for (fdataset in datasets) {
##     for (fcomp in comps) {
##         for (fmerge_method in merge_methods) {
##             for (fprob in probs) {
##                 ## algo = algos[1]
##                 max_k_here = max_changed_k %>%
##                     filter(dataset == fdataset,
##                            comp == fcomp,
##                            merge_method == fmerge_method,
##                            problem == fprob)
                
##                 if (nrow(max_k_here) == 0) next

##                 print(max_k_here$max_changed_k)
                                    
##                 plot_data <- tidied %>%
##                     filter(dataset == fdataset,
##                            comp == fcomp,
##                            merge_method == fmerge_method,
##                            problem == fprob,
##                            k > 0,
##                            k <= max_k_here$max_changed_k)

##                 plotname = paste("vs-k", fdataset, fcomp, fmerge_method, fprob, ".png", sep = "-")
##                 print(paste("Generating ", plotname))
##                 png(filename = plotname)
                
##                 p = ggplot(plot_data, aes(x = as.factor(k), y = as.numeric(hindex_increase))) +
##                     geom_boxplot(aes(fill = measure)) +
##                     labs(
##                         x = "Number of allowed operations",
##                         y = "h-index increase",
##                         fill = "Measure")
##                 print(p)
##                 dev.off()
##             }
##         }
##     }
## }

## Facet plot dataset vs unmerge algorithm (problem) for compat
## threshold 0.4 and greedy_maxdeg merges.

dat_plot <- filter(tidied, comp == 0.4, merge_method == "GreedyMax")

max_k_summary <- max_changed_k %>%
    ungroup %>%
    filter(comp == 0.4, merge_method == "GreedyMax") %>%
    summarize(max_changed_k = max(max_changed_k))

print(max_k_summary)

dat_plot <- filter(dat_plot, k > 0, k <= max_k_summary$max_changed_k)

tikz("vs-k-facet-comp-0.4-greedy_maxdeg.tex",
     pointsize = 11,
     width = 4.8,
     height = 6)

p <- ggplot(dat_plot) +
    geom_boxplot(coef = 0,
                 aes(x = as.factor(k),
                     y = as.numeric(hindex_increase),
                     fill = measure),
                     ## color = measure),
                 outlier.size = 0.3) +
    ## guides(color = FALSE) +             # Remove legend for color aesthetic.
    labs(x = "Number of changed articles or operations",
         y = "H-index increase",
         fill = "Measure") +
    theme(legend.title = element_blank(),
          legend.position = c(0.5, 0.905)) + 
    facet_grid(problem ~ dataset)

print(p)
dev.off()

