'''

Split Index: Maximizing h-Indices of Google Scholar profiles by
unmerging merged articles.

Copyright (C) 2016-2019 Clemens Hoffmann, Manuel Sorge, and Kolja
Stahl

For licensing see LICENSE.

'''

import sys

import networkx as nx

from matplotlib import pylab
import matplotlib.pyplot as plt


def log(msg):
    sys.stderr.write(msg + '\n')
    sys.stderr.flush()


def parse_id(row):
    """Parse the cluster id of an article from a row in the raw data
    generated by scholar.py. If the cluster id is not available, use
    the article id.
    """

    cid = None
    aid = row[10]
    if row[5] != 'None':
        cid = row[5]
    elif row[7] != 'None':
        log('No cluster id in clusterid field present for %s, '
            'using citation-list link field' % "|".join(row))
        cid = (((row[7].split('='))[1]).split('&'))[0]
        # print(cid)
    elif row[8] != 'None':
        log('No cluster id present in clusterid or citation-list field '
            'for %s, using versions-list link field' % "|".join(row))
        cid = (((row[8].split('='))[1]).split('&'))[0]
        # print(cid)            
    else:
        log('Did not find cluster id in clusterid, citation-list, '
            'or versions-list fields for %s, using article id' % "|".join(row))
        cid = aid

    return cid


def read_cite_graph(cfilename):
    D = nx.DiGraph()
    with open(cfilename + ".gra", 'r') as graphfile:
        for line in graphfile:
            tabs = line.split()
            a = tabs[0]
            b = tabs[1].strip()
            D.add_edge(b, a)

    with open(cfilename + ".articles") as articlefile:
        for line in articlefile:
            tabs = line.split()
            D.add_node(tabs[0], purged=int(tabs[1]))

    return D


def read_comp_graph(path, th):
    G = nx.Graph()
    with open(path + "/" + th + ".comp") as graphfile:
        for line in graphfile:
            tabs = line.split()
            a = tabs[0]
            b = tabs[1].strip()
            G.add_edge(a, b)

    with open(path + ".articles") as articlefile:
        for line in articlefile:
            tabs = line.split()
            G.add_node(tabs[0])

    return G


def draw_graph(graph, file_name):
    # initialize Figure
    plt.figure(num=None, figsize=(20, 20), dpi=80)
    plt.axis('off')
    fig = plt.figure(1)
    pos = nx.spring_layout(graph)
    nx.draw_networkx_nodes(graph, pos)
    nx.draw_networkx_edges(graph, pos)
    nx.draw_networkx_labels(graph, pos)

    cut = 1.20
    xmax = cut * max(xx for xx, yy in pos.values())
    ymax = cut * max(yy for xx, yy in pos.values())
    plt.xlim((1 - cut) * xmax, xmax)
    plt.ylim((1 - cut) * ymax, ymax)

    plt.savefig(file_name, bbox_inches="tight")
    pylab.close()
    del fig
